/*
   ----------------------------------------------------------------------------
   TempTableReport2.prg
   Sample code from The dBASE Reports Book, second edition
   Chapter 11
   Copyright  2016, Kenneth J. Mayer
   
   Program code to create a temporary table using the "REPnnnn" filename
   template, and generate a report from that. This is the second phase --
   completed.
   
   This may be used for other purposes, as needed. 
   ----------------------------------------------------------------------------
*/
// Create a temporary table for a report
clear

// set up a database object:
local oData
oData = new database()
oData.databaseName := "DBASEREPORTS"
oData.active := true

/*
   ***********************************************************
   If you are using dBASE Plus 8 or later, we do not need the
   code below for the BDE Alias to obtain the path to the 
   folder referenced by the alias. Instead we can use the 
   Database Alias created back in Chapter 3 ...
   ***********************************************************
*/
nVersion = __version__
if nVersion < 8

   // set up a BDE Alias object briefly so we can
   // get the path to the folder with the tables -
   // this is necessary for the FUNIQUE() function
   // to work properly.
   set procedure to :dUFLP:BDEAlias.cc
   local oBDE, cPath
   // create an instance of the object:
   oBDE = new BDEAlias()
   // get the path to the folder:
   cPath = oBDE.databaseDir( "DBASEREPORTS" )
   // clean up memory:
   release object oBDE
   oBDE := null
   close procedure :dUFLP:BDEAlias.cc

else  // current versions of dBASE

   // in 'current' versions of dBASE, using the User BDE Alias,
   // the path is stored in the PLUS.INI file, so we need to read
   // the data from there ... we'll use the INI.cc file in the
   // dUFLP instead:
   set procedure to :dUFLP:INI.cc
   local oINI, cPath
   // instance of the INI object with the path to the 
   // file from the app object in dBASE
   oINI = new ini( _app.iniFile )
   // find the value we need:
   cPath = oINI.GetValue( "dBASEReports", "Options" )
   // remove "PATH:" from beginning
   cPath = right( cPath, len( cPath ) - 5 ) 
   // cleanup
   release object oINI
   oINI := null
   close procedure :dUFLP:INI.cc
endif

// The path returned does not include a trailing
// backslash character:
cPath += "\"

// The FUNIQUE function is used to create a unique
// filename, using wildcards for the characters you
// need. For example, here we.re going to call the
// file: "REPnnnnn.dbf" where the first three letters
// are "REP", the "nnnnn" will be randomly created
// by the function, and the file extension is specified
// as ".dbf". What we need to do is make sure that the
// filename as returned by FUNIQUE does not exist:
local cFile
do while true // infinite loop
   // create a filename:
   cFile = funique( cPath+"REP?????.dbf" )
   // if the table does not exist, exit the loop
   if not oData.tableExists( cFile )
      exit
   endif
enddo

// if we're here we have a unique filename.
// we must create the table. To do this we will
// build the command as a character string, and
// then use the database object to create the table:
local cTable
// get just the table name from the path/file string:
cTable = right( cFile, len( cFile ) - rat( "\", cFile ) )

// and remove the ".dbf" at the end, because
// the CREATE TABLE command will create a file
// called "dbf" if that's there, which is not
// what we want:
cTable = left( cTable, at( ".", cTable ) -1 )

// tell the data object to copy the template table
// to the temporary table name:
oData.copyTable( "ReportTemplate", cTable )

// now we have the temporary table, we can put
// data into it. This requires setting up
// the query objects for the tables:
local qDams, qDogs, qTemp
qDams = new query()
qDams.database := oData
qDams.sql := "select * from dams"
qDams.active := true
qDams.rowset.indexName := "NAME"

qDogs = new query()
qDogs.database := oData
qDogs.sql := "select * from dogs"
qDogs.active := true
qDogs.rowset.indexName := "DAM"
qDogs.rowset.masterRowset := qDams.rowset
qDogs.rowset.masterFields := "DAM"

qTemp = new query()
qTemp.database := oData
qTemp.sql := "select * from "+cTable
qTemp.active := true

// now that we have the query objects set,
// we need to loop through the data, and
// add rows to the temporary table:
qDams.rowset.first()
do while not qDams.rowset.endOfSet
   // output so we know *something* is happening:
   ?
   ? qDams.rowset.fields["NAME"].value
   
   // inner loop for the child rowset - dogs:
   qDogs.rowset.first()
   do while not qDogs.rowset.endOfSet
      // output:
      ? " "+qDogs.rowset.fields["NAME"].value
      
      // for each row in the DOGS table we need
      // a row in the temp table:
      qTemp.rowset.beginAppend()
      // name of the dam:
      qTemp.rowset.fields["DamName"].value := ;
              qDams.rowset.fields["NAME"].value
      // name of the child/dog:
      qTemp.rowset.fields["ChildName"].value := ;
              qDogs.rowset.fields["NAME"].value
      // birthdate of the dog:
      qTemp.rowset.fields["BirthDate"].value := ;
              qDogs.rowset.fields["BORN"].value
      // make sure we save this!
      qTemp.rowset.save()
      // next row in the Dogs table:
      qDogs.rowset.next()
   enddo
   
   // next row in the Dams table:
   qDams.rowset.next()
enddo

// output:
?
? "** Done building temporary table!"

// we're done building the temporary table!
// code will be added here to generate
// and display the report ...

set procedure to TemporaryTable.rep additive
local oRep
oRep = new TemporaryTableReport()

// we have an instance of the report, now
// let's change the Query1 object's SQL property:
oRep.Query1.sql := "select * from "+cTable

// at this point, if we render this, the temporary
// table will be deleted, perhaps before we want it
// to. Instead, we will use the Preview form from
// the dUFLP (as shown in the previous chapter):
set procedure to :dUFLP:Preview.wfm additive
oPF = new PreviewForm()
oPF.Viewer.ref := oRep
oPF.bModal = true
oPF.readModal()

// when this screen is closed, then we can
// move on and do our cleanup:
oPF.release()
close procedure :dUFLP:Preview.wfm
oRep.release()

// cleanup -- we want to delete the temporary table,
// but to do that we must remove references to it:
qTemp.active := false
release object qTemp
qTemp = null
// delete the table:
oData.dropTable( cTable )

// remove references to the other tables:
qDogs.active := false
release object qDogs
qDogs = null
qDams.active := false
release object qDams
qDams = null

// and finally get rid of the database object:
oData.active := false
release object oData
oData = null
