/*
SerialComm.cc

-----------------------------------------------------------------------------
Description:

   Visual component class that sets up and manages a serial communication
   handler for Visual dBase 7.01 applications.

Programmer:

   Nicolas Martin, nmartin@magic.fr

Copyright notice:

   This file is owned by Nicolas Martin. See the file "Copyright.txt" for full
   copyright details and terms of use.

History:

   10/19/2002 -- Modified by Jean-Pierre Martel
                 The Custom properties and some Functions from Nicolas Martin's original version
                 were put back in order to work with the demo form TTYDemo.wfm
   12/22/2001 -- Modified by Vic McClung to work with db2K x.x
                 This problem was caused by a bug fix in dB2K.  Nicholas was
                 using the bug. i.e. using a reference to this as meaning the
                 SerialComm object while in the timer.OnTimer event, when properly
                 this within that event handler must refer to the timer object.
                 I have hooked my desktop and notebook together with a serial cable
                 and null modem and tested all of the demo forms furnished by Nicolas
                 with this class and also tested the modem parts of it on my notebook
                 which has a serial port modem.  All seemed to work properly without
                 modification.
   12/07/1998 -- Original V1
   06/10/1998 -- V 1.01 :
                Clarify comments, add minor improvements, correct minor bugs :
                Improvements :
                + The call to the user ISR when the continuous read is
                  activated is now performed by a try/catch exception
                  handler, for better robustness.
                + The fullDuplexWrite Function has been simplified and
                  enhanced, based on a different way to process the read
                  operation while sending the data.
                + flushPort Function added, performed by a Win32 API call
                  that completly flushes the com port (read _&_ write).
                + Added a version property to show within the inspector
                  or the application the current version of the component.
                Bugs fixed :
                + The return code of several Functions was incorrect when
                  the com port was not opened (returned true instead of false)
                + The continuousRead Function has been changed in order to
                  empty the com port read buffer at each tick, to avoid
                  loss of received data due to timer jitter and minimum value.
                  This leads to a slightly different behaviour of that Function.

Dependencies : StructAPI.h, StructAPI.prg, StrucMem.dll, StructureEx.cc

Designed for Visual dBase, version 7.01

How to use:

   Add the SerialComm.cc component to your Visual dBase component palette
   (See your Visual dBase 7 User Manual if you need help to do that, or consult
   the online help, search for "custom components:adding to palette").
   You can then select the SerialComm component from the Custom page of the
   component palette and simply add it to your form in order to access to the
   component methods. The component appear on the form in design mode, it does
   not appear in run mode, (it behaves somehow like the Query component) so you
   can deposit it anywhere on your form, even above other objects.
   Or you can choose to create an instance of the SerialComm object by programmation,
   it would look like :

          set procedure to SerialComm.cc additive
          this.serialComm1 = new SerialComm( this )

Properties:

   displayMessages
      Flag, indicates whether SerialComm displays internal messages when errors
      occur during execution. displayMessages defaults to true. When setting
      displayMessages property to false, no error messages will be displayed.
      However, testing the return codes of each Method allows you to know whether
      an error occured during its execution.

   version
      String, shows the current version of the serialComm component. This property
      can be seen in the inspector (in the form design mode, for instance) or accessed
      within your program.

Methods:

   openPort( <ComPort>, <BaudRate>, <parity>, <data>, <stopBits> )
      This Function opens and initializes the Comm port you want to access.
      By default, the handsaking method is set to "no handshake".
      The parameters are:
         <ComPort>: VdB string, containing the DOS name "COM1", "COM2", "COM3", ...
         <BaudRate>: VdB Numeric, communication line speed.
         <parity>: VdB string, can be "None", "Even", "Odd", or simply 'N', 'E', or 'O'
         <data>: VdB numeric, number of bits per characters
         <stopBits>: VdB numeric, can be : 1 or 1.5 (check your computer!), or 2
      The Function returns true if the port has been opened, and false otherwise

   closePort( )
      This Function closes a previously opened communication port. No parameter
      is required, the Comm port that has been opened by the current object will
      be closed.
      The Function returns true if the port has been closed, or false otherwise

   readString( <MaxChars> )
      One shot read.
      Reads the data that has already been received on the communication port.
      If no data is available, the Function returns with an empty string.
      If data is present, it is placed in the returned string.
          <MaxChars> : VdB numeric, maximum number of characters to read. The
                       Function will read a number of characters smaller or equal
                       to <MaxChars>.
      The Function will return a string with the data read, and with length equal
      to the numbers of characters read, or an empty string.

   writeString( <Wstring> )
      Sends the data in <Wstring> on the communication line. The Function will
      send all the data contained in Wstring.
          <Wstring>: VdB string, the buffer of characters to send.
      The Function returns true after all the characters have been sent, and false otherwise

   setHandshake( <HSmethod> )
      Sets the Handshaking method for the communication flow control, for read
      and write operations.
          <HSmethod>: VdB numeric, handshaking method, can be 0 for no handshake,
                      1 for hardware (DTR/CTS) handshake, and 2 for Xon/Xoff handshake.
      The Function returns true if it has been correctly executed, or false otherwise

   startContinuousRead( <bufferSize>, <callBack> )
      This high level Function has been designed in order to perform Full Duplex
      communications with very little programmatic effort.
      The goal is to set the SerialComm component in reception mode so that
      it behaves somehow like an Interrupt handler, notifying a user Interrupt
      Service Routine (ISR) as soon as new characters have been received on the
      reception line. Actually, after the startContinuousRead Function has been
      executed, the following activity is being initialized :
      - A VdB timer is set up, and fires regularly to look whether new
        characters have been received on the reception line.
      - If new characters are available, the new characters are placed in
        a string and passed as a parameter to the user ISR (callBack routine)
        which is then called. The way it performs is closer to a Multi-Thread
        application than to an Interrupt Handler, but the result is roughly the same,
        with some particularities explained below.
        The parameters passed to startContinuousRead are :
           <bufferSize> : VdB numeric, specifies the desired maximum number of characters
                          that should be passed in the parameter string to the ISR.
                          In fact, more or less characters can be passed each time to the
                          ISR, depending on the count actually received, and on the timer
                         jitter.
           <callBack> :  VdB Function pointer, of the ISR that is in charge
                         of processing the characters read from the port.
                         This ISR must be declared with a parameter, which is the
                         name of the string that contains the new characters received.
                         For instance, if the ISR is part of a form, it can be
                         declared as :

                         Function processReadData( newChars )
                            ...
                            process the characters in newChars
                            ...
                           return

                         to initialize the read thread, you can place a :

                            this.serialComm1.startContinuousRead( 50, this.processReadData )

                         in the form's onOpen method, to get typically 50 chars in newChars.

        The Function returns true if it has been correctly initialized, or false otherwise

        Important Remarks and Recommandations
       -------------------------------------
       1/ Serial communication line arithmetics
       As described above, the read thread uses a timer to check for the presence
       of new characters. The timer tick is set to a value corresponding to the
       minimum time to get a number of characters equal to <bufferSize>, given
       the parameters used to open the port. This value (in seconds) is :

       <bufferSize> * ( 1 + <data> + <parityBit> + <stopBits> ) / <BaudRate>

       Note: a character always contains 1 start bit, additionnaly to the
       data, the parity and stop bits.

       In order to get efficient software, you need to set a <bufferSize> not
       too small when <BaudRate> gets high. Furthermore, your user-defined ISR
       execution time must be smaller than the timer value.
       For instance, with :
       <bufferSize> = 150, <data> = 8, <stopBits> = 1, <BaudRate> = 19200, <Parity> = 'None'
       the timer tick will be set to 78 milliseconds.

       IMPORTANT : the minimum timer tick handled by VdB is 55 ms. If the timer
       ---------   tick calculated above in the startContinuousRead Function is
                   lower, the minimum value of 55 ms will used by VdB.
                   In all the cases, the port will be read as much as necessary to
                   empty it. This may lead to having always more characters
                   than <bufferSize> at each ISR call.

       2/ Full duplex recommandation
       The writeString operation is performed in one block, this means that the
       writeString Function will not be interrupted until all the chars have
       been sent on the port.
       Thus, when performing Full Duplex communications, you should not send
       a long string in one block with writeString, but rather use the Function
       fullDuplexWrite, which enables read and write at the same time.


   stopContinuousRead( )
      After a startContinuousRead, the acquisition goes on until either :
      - a stopContinuousRead is performed,
      - or the Comm port is closed.
      The Function returns true if it has been correctly executed, or false otherwise

   fullDuplexWrite( <WString> )
       This Function is intented to be used when Full Duplex communications are
       to be handled. In that case, the fullDuplexWrite Function performs the
       transmission of the characters of <WString>, alternatively with
       'readString' on the reception line when continuous read has been previously
       enabled. When characters are available in reception, the user ISR is called
       the same way as defined for 'startContinuousRead'.
            <WString> : VdB string, contains the characters to be sent.
      The Function returns true after all the characters have been sent, and false otherwise


   sendBreak( <duration> )
      This enables the user to send a communication BREAK on the serial line.
      A communication BREAK is typically sent in order to stop the communication
      process by either one of the two devices that are at each end of the line.
      The duration of the BREAK command can be specified.
         <duration> : VdB numeric, specifies the duration (in seconds) of the BREAK.
                      If no parameter is passed to this Function, a default BREAK
                      of 1 second is sent on the line.
      The Function returns true if it has been correctly executed, or false otherwise

   flushPort( )
      This Function flushes the read and write buffers from any characters.
      The Function returns true if it has been correctly executed, or false otherwise

   flushReadPort( )
      This Function flushes the reception line from any characters.
      The Function returns true if it has been correctly executed, or false otherwise

   Miscellaneous:
      Want to learn more about RS232 serial communication on PC computers ?
      visit:	http://www.ridgewater.mnscu.edu/classes/dc/io/index.htm
-----------------------------------------------------------------------------
Here starts the code :
*/

class SerialComm( f ) of Rectangle( f ) custom
// the following protections makes properties and methods hidden to the user.
		protect ;
		printable, visible, ;
		fontBold, fontItalic, fontName, fontSize, fontStrikeout, fontUnderline, ;
		text, ;
		border, borderStyle, colorNormal, mousePointer, pageNo, patternStyle, ;
		onDesignOpen, onLeftDblClick, onLeftMouseDown, onLeftMouseUp, ;
		onMiddleDblClick, onMiddleMouseDown, onMiddleMouseUp, onMouseMove, onOpen, ;
		onRightDblClick, onRightMouseDown, onRightMouseUp, ;
		chk, serialOnClose, serialCommOnOpen, timerLoop, nativeSessionTimerLoop, lpMess, ;
		errorPortNotOpened, dcb, bufferSize, byteBuffer, callBack, timer1, ComStat


// include VdB files
#include "winbase.h"
#include "winnt.h"
#include "Vdbase.h"

// defines
#define FILE_NOT_SHARED	0
#define MINUS_ONE			0xffffffff
#define BIGBUFFERSIZE	10000
#define ERRORTITLE		'Error in SerialComm'

	// configure this object

	with ( this )
		onOpen := this.SerialCommOnOpen
		text := ' SerialComm '
		height := 2
		width := 14
		visible := false
		colorNormal := 'yellow/navy'
	endwith

	// set the custom properties
	this.displayMessages = true						// default set to true, display internal messages 
	this.version = 'SerialComm V 1.5'				// version identification
	this.BaudRate = Null									// BaudRate
	this.Data = Null										// Number of data bits
	this.ComPort = Null          						// DOS name for the com port
	this.Parity = Null									// Parity, if any
	this.StopBits = Null									// Number of stop bits
	this.CommErrors = new CommErrorClass( )		// contains info about the last errors detected
	this.PortStatus = new PortStatusClass( )		// contains info about the port status

	// reference to ComPort
	this.hSerial = MINUS_ONE

	// include declarative structure classes for the StructureEx.cc class
	private sFolder
	sFolder = program( 0 )
	sFolder = left( sFolder, sFolder.lastIndexOf( '\' ))
	set procedure to "&sFolder.\StructureEx.cc" additive

   this.ComStat = new _COMSTAT()				   // Creates a ComStat structure
	this.dcb = new _DCB()                		// Creates a DCB structure
	this.CommProp = new _COMMPROP()          	// creates a COMMPROP structure

	// loads the API 32 Functions used by SerialComm
	extern HANDLE CreateFile(LPCSTR, DWORD, DWORD, LPSTRUCTURE, DWORD, ;
						DWORD, HANDLE) kernel32 from "CreateFileA"
	extern BOOL   CloseHandle(HANDLE) kernel32
	extern BOOL   GetCommState(HANDLE, LPSTRUCTURE) kernel32
	extern BOOL   SetCommState(HANDLE, LPSTRUCTURE) kernel32
	extern BOOL   BuildCommDCB(LPCSTR, LPSTRUCTURE) kernel32 from "BuildCommDCBA"
	extern BOOL   WriteFile(HANDLE, LPCVOID, DWORD, LPDWORD,LPSTRUCTURE) kernel32
	extern BOOL   ReadFile(HANDLE, LPVOID, DWORD, LPDWORD,LPSTRUCTURE) kernel32
	extern BOOL   GetCommTimeouts(HANDLE, LPSTRUCTURE) kernel32
	extern BOOL   SetCommTimeouts(HANDLE, LPSTRUCTURE) kernel32
	extern BOOL   ClearCommError(HANDLE, LPDWORD, LPSTRUCTURE) kernel32
	extern BOOL   SetCommBreak(HANDLE) kernel32
	extern BOOL   ClearCommBreak(HANDLE) kernel32
	extern DWORD  GetLastError(VOID) kernel32
	extern BOOL   PurgeComm(HANDLE, DWORD) kernel32
	extern BOOL   GetCommProperties(HANDLE, LPSTRUCTURE) kernel32
   extern BOOL   SetupComm(HANDLE, DWORD, DWORD) kernel32
	extern BOOL   GetCommMask(HANDLE, LPDWORD) kernel32
	extern BOOL   SetCommMask(HANDLE, DWORD) kernel32

	Function serialCommOnOpen
   	// Linked method : Serial Comm On open event
		// the paintbox created here tracks an OnClose event, for checking whether
		// the Com port has been properly closed before closing Serial Comm.
		// These lines can be safely commented to desactivate this capability.
		this.PaintBox1 = new paintbox( this.parent )
		with ( this.PaintBox1 )
			transparent := true
			visible := false
			onClose = this.serialOnCLose
		endwith
		this.PaintBox1.SerialCommRef = this		// keep track of serialComm object ref
		return

	Function openPort( ComPort, BaudRate, parity, data, stopBits )
   	// this Function opens the communication port
		local commStr, commTO

		// Create the timer for the continuous read Function
		this.timer1 = new timer( )

		// Opening a communication port is performed in 4 steps :
		// 1. First, we need to create the file handle to the communication port
		this.hSerial = CreateFile( ComPort ,;
									GENERIC_READ + GENERIC_WRITE ,;
									FILE_NOT_SHARED ,;
									Null ,;
									OPEN_EXISTING ,;
									FILE_ATTRIBUTE_NORMAL ,;
									Null )		// opens communication port

		if ( this.hSerial == MINUS_ONE )	// Checks it is opened correctly
			if this.displayMessages
				AlertMessage( 'Unable to open COM port', ERRORTITLE )
			endif
			return false		// Abort OpenPort
		endif

		// 2. Now, we set the communication port parameters
		this.dcb = replicate( chr( 0 ), 32 )		// Creates a DCB
		if not this.chk( GetCommState( this.hSerial, this.dcb ))  // Gets DCB values
			return false
		endif

		// builds the Comm String using the classic DOS "mode" command
		CommStr = 'baud='+ BaudRate + ' parity=' + parity.left( 1 )
		CommStr += ' data=' + data + ' stop=' + stopBits
		CommStr += chr( 0 )  // and do not omit chr( 0 ), otherwise, it does not work !

		if not this.chk( BuildCommDCB( CommStr, this.dcb ))
			return false
		endif
		if not this.chk( SetCommState( this.hSerial, this.dcb ))  // Sets DCB values
			return false
		endif

		// 3. Here, we set a default NO Handshake for the Handshaking method.
		this.setHandshake( 0 )

		// 4. Last, we build the timeouts for read operation.
		// ..."A value of MAXDWORD, combined with zero values for both the
		// ReadTotalTimeoutConstant and ReadTotalTimeoutMultiplier members,
		// specifies that the read operation is to return immediately with
		// the characters that have already been received, even if no characters
		// have been received"...
		commTO = replicate( chr( 0 ), 21 )	// builds COMMTIMEOUTS
		commTO.setbyte( 0, 0xff )
		commTO.setbyte( 1, 0xff )
		commTO.setbyte( 2, 0xff )
		commTO.setbyte( 3, 0xff )

		// Update the object properties with the port configuration parameters.
		with ( this )
			BaudRate = BaudRate
			Data = Data
			ComPort = UPPER(ComPort)
			Parity = Proper(parity)
			StopBits = StopBits
		endwith

		if not this.chk( SetCommTimeouts( this.hSerial, commTO )) // Sets CommTimeouts
			return false
		endif

		// Compute the acquisition time per character, for continuous read timer tick
		this.secondPerChar = ( 1 + data + iif( parity.left( 1 ) == 'N', 0, 1 ) + stopBits ) / BaudRate

		// these lines can be uncommented to check the DCB values, for debug purpose.
//		for i = 0 to this.dcb.length - 1
//			? itoh(this.dcb.getbyte(i))
//		endfor

		return true

	Function closePort
   	// this Function closes the previously opened communication port
		if ( this.hSerial <> MINUS_ONE )
			this.stopContinuousRead( )
			if not this.chk( CloseHandle( this.hSerial ))
				return false
			endif
			this.hSerial = MINUS_ONE
		else
			this.errorPortNotOpened( )
			return false
		endif
		return true

	Function setHandshake( HSmethod )
   	// this Function sets the HandShaking method
		if ( this.hSerial == MINUS_ONE )
			this.errorPortNotOpened( )
			return false
		endif
		do case
			case ( HSmethod == 0 )		// No handshake method
				// ..."fInX, fOutX,fOutXDsrFlow,and fOutXCtsFlow are all set to FALSE
				//fDtrControl is set to DTR_CONTROL_ENABLE
				//fRtsControl is set to RTS_CONTROL_ENABLE"...
				this.dcb.setByte( 8, 0x13 )
				this.dcb.setByte( 9, 0x10 )
			case ( HSmethod == 1 )		// Hardware handshake
				// ..."fInX, fOutX are both set to TRUE
				// fOutXDsrFlow,fOutXCtsFlow are both set to FALSE.
				// fDtrControl is set to DTR_CONTROL_ENABLE
				// fRtsControl is set to RTS_CONTROL_ENABLE"...
				this.dcb.setByte( 8, 0x13 )
				this.dcb.setByte( 9, 0x13 )
			case ( HSmethod == 2 )		// Xon/Xoff handshake
				// ..."fInX, fOutX are both set to FALSE
				// fOutXDsrFlow,fOutXCtsFlow are both set to TRUE.
				// fDtrControl is set to DTR_CONTROL_HANDSHAKE
				// fRtsControl is set to RTS_CONTROL_HANDSHAKE"...
				this.dcb.setByte( 8, 0x2f )
				this.dcb.setByte( 9, 0x20 )
			otherwise
				if this.displayMessages
					AlertMessage( 'Bad handshake method requested. can be 0,1 or 2 only.', ERRORTITLE )
				endif
				return false
		endcase
		if not this.chk( SetCommState( this.hSerial, this.dcb ))  // Sets DCB values
			return false
		endif
		return true


	Function readString( MaxChars )
   	// This Function performs reads data from the communication port.
	   // The trick here is that we need to convert the byte string returned by ReadFile, into
   	// a VdB Unicode string which is stored as a 16 bits word string, one character in each word.
		local i, buffer, sRBuffer, nBytesRead
		nBytesRead = 0		// will be set by ReadFile Function
		if ( this.hSerial <> MINUS_ONE )
			buffer = replicate( chr( 0 ), MaxChars )	// creates the byte buffer
			if not this.chk( ReadFile( this.hSerial, buffer, MaxChars, nBytesRead, Null ))
				return ''
			endif
			sRBuffer = ''		// Now, convert the byte string into a VdB standard string

         for i = 0 to nBytesRead - 1
            if buffer.getByte(i) = 0
               sRbuffer += "0" 											// Patch to collect a "NULL" transmission
                   else
               sRBuffer += chr( buffer.getByte( i ))
         endif
         endfor
			return sRBuffer
		else
			this.errorPortNotOpened( )
		endif
		return ''

	Function startContinuousRead( bufferSize, callBack )
   	// this Function starts the continuous read operation.
		if ( this.hSerial <> MINUS_ONE )
			// creates the buffer for continuous read operations
			this.byteBuffer = replicate( chr( 0 ), bufferSize )	// creates the byte buffer
			this.bufferSize = bufferSize  // # of chars max. to receive
			this.callBack = callBack		// saves callback routine name
			this.timer1.interval := this.secondPerChar * bufferSize
         //this.timer1.NSTL = this.nativeSessionTimerLoop
         this.timer1.onTimer := class::nativeSessionTimerLoop // this.timerLoop
         // this code added by Vic McClung to overcome problems caused by
         // fixing bug in Visual dBASE 7.01
         this.timer1.byteBuffer = this.byteBuffer
         this.timer1.bufferSize = this.bufferSize
         this.timer1.hSerial = this.hSerial
         this.timer1.callBack = this.callBack
         this.timer1.parent = this
         this.timer1.form = this.parent
         this.timer1.displayMessages = this.displayMessages
         // end of code added
			this.timer1.enabled := true		// starts the read thread.
		else
			this.errorPortNotOpened( )
			return false
		endif
		return true

   //Function timerLoop
	//    this Function is used as the continuous read timer loop
   //   return this.NSTL( )

	Function nativeSessionTimerLoop
   	// this Function has been created in order to execute the timer loop
	   // steps inside the original form's session. This has been added to
	   // keep the form's session settings during loop execution, and especially
	   // for the SET TALK primitive
      // Changed all references from this. to class:: for class methods
      // Changed by Vic McClung because of problem mentioned above.
      // In this method, this refers to the timer and not SerialComm
		local i, sRBuffer, nBytesRead
		sRBuffer = ''
		nBytesRead = 0		// will be set by ReadFile Function
		do    // read as much as needed to empty read port buffer
         if not class::chk( ReadFile( this.hSerial, this.byteBuffer, this.bufferSize, nBytesRead, Null ))
				return false
			endif
			// Now, convert the byte string into a VdB standard string
			for i = 0 to nBytesRead - 1
				sRBuffer += chr( this.byteBuffer.getByte( i ))
			endfor
		until ( nBytesRead < this.bufferSize )
		if sRBuffer.length > 0	// New characters read ?
			try	// try to call the user ISR
            this.callBack( sRBuffer )
			catch ( Exception e1 )
            class::stopContinuousRead( ) // aborts continuous read
            if this.displayMessages
					AlertMessage( 'Invalid user ISR Function pointer', ERRORTITLE )
				endif
				return false
			endtry
		endif
		return true

	Function stopContinuousRead
   	// this Function stops the continuous read operation.
      local o
		if ( this.hSerial <> MINUS_ONE )
         o = this
         if o.ClassName = "TIMER"
            this.enabled := false
         else
            this.timer1.enabled := false
         endif
		else
         class::errorPortNotOpened( )
			return false
		endif
		return true

	Function writeString( Wstring )
   	// This Function writes a VdB string to the communication port.
	   // The same trick applies here as for the read operation, convertion of the VdB
   	// Unicode string to send, stored as a 16 bits word string with one character in each word,
	   // into a byte string containing the characters to send.
		if this.hSerial <> MINUS_ONE
			local i, bytesWritten, sendStr
			bytesWritten = 0		// will be set by WriteFile
			sendStr = replicate( chr( 0 ), Wstring.length )  // creates byte string
			for i = 0 to Wstring.length - 1				// converts VdB word string to byte string
				sendStr.setByte( i , asc( Wstring.charAt( i )))
			endfor
			if not this.chk( WriteFile( this.hSerial, sendStr, Wstring.length, bytesWritten, Null ))
				return false
			endif
			if bytesWritten <> Wstring.length		// checks # of bytes written
				if this.displayMessages
					AlertMessage( 'Error in writeString Function', ERRORTITLE )
				endif
				return false
			endif
		else
			this.errorPortNotOpened( )
			return false
		endif
		return true


	Function fullDuplexWrite( Wstring )
   	// This Function it typically used to send data in Full Duplex operation.
	   // The string Wstring to be sent is splited and sent in groups of characters, each
	   // group containing the number of chars specified at the startContinuousRead.
	   // Between each group transmission, the reception line is checked (if continuous
	   // read has previously been set ) and user ISR called if new chars are arrived.
		local j, fContReadStart, bufferSize
		if ( this.hSerial <> MINUS_ONE )
			fContReadStart = this.timer1.enabled			// records that continuous read is started
			bufferSize = iif( fContReadStart, this.bufferSize, Wstring.length	)	// if not read continuous started, send string in 1 block
			this.timer1.enabled := false		// stops timer ticks and performs polling read
			// processes the string, sending each time bufferSize characters
			for j = 0 to Wstring.length - 1 step bufferSize
				this.writeString( Wstring.subString( j, j + bufferSize ))
				if fContReadStart				// continuous read was started ?
					this.nativeSessionTimerLoop( )	// read port and process user ISR
				endif
			endfor
			this.timer1.enabled := fContReadStart		// restarts timer ticks, if continuous read
		else
			this.errorPortNotOpened( )
			return false
		endif
		return true


	Function sendBreak( duration )
   	// Sends a communication BREAK on the line. Default duration will be
	   // one second, when not specified.
		local nd
		if ( this.hSerial <> MINUS_ONE )
			nd = iif( argcount( ) == 0, 1, duration )
			if not this.chk( SetCommBreak( this.hSerial ) )
				return false
			endif
			sleep nd
			if not this.chk( ClearCommBreak( this.hSerial ) )
				return false
			endif
		else
			this.errorPortNotOpened( )
			return false
		endif
		return true

	Function flushPort
   	// Flushes the communication port, read and write.
		if ( this.hSerial <> MINUS_ONE )
			if not this.chk( PurgeComm( this.hSerial, PURGE_TXABORT + ;
																	PURGE_RXABORT + ;
																	PURGE_TXCLEAR + ;
																	PURGE_RXCLEAR ))
				return false
			endif
		else
			this.errorPortNotOpened( )
			return false
		endif
		return true

	Function flushReadPort
   	// Flushes the communication port, read line.
		// When port is flushed, readstring will return ''
		do while this.readString( BIGBUFFERSIZE ).length > 0
		enddo
		return true

	Function chk( stat )
   	// this Function checks the return value of the Win 32 API calls.
		if not stat
			if this.displayMessages
				AlertMessage( 'Error in Win 32 API return call', ERRORTITLE )
			endif
			return false
		endif
		return true

	Function serialOnClose
   	// this Function, linked to the PaintBox onClose event, checks whether the
	   // user has properly closed the comm Port before closing the container or the form.
	   // If not, it asks for that.
	   // Be aware that unclosing the comm port will make it unaccessible after. You
	   // will need to restart either VdB or your Compiled .EXE application.
		if ( this.serialCommRef.hSerial <> MINUS_ONE )
			this.serialCommRef.stopContinuousRead( )		// stop continuous read in any case
			if this.serialCommRef.displayMessages
				this.ret = ConfirmationMessage( 'The Comm port has not been closed. Do you want to close it now ?' ,;
								 'SerialComm Warning !' )
				if ( this.ret == BUTTON_YES )
					this.serialCommRef.closePort( )
				else
					return false
				endif
			else
				return false
			endif
		endif
		return true

	Function errorPortNotOpened
   	// this Function displays a message after trying to execute a Function while the
	   // com port is not yet opened. The message is displayed if the displayMessages
	   // property of the component is set to true
		if this.displayMessages
			AlertMessage( 'Cannot execute Function, the Com Port is not opened', ERRORTITLE )
		endif
		return

	Function setXonXoffConfig( nXonChar, nXoffChar, nXonPct, nXoffPct, nInBuffSize )
   	// This function sets the Xon/Xoff chars, and optinnaly the Xon and Xoff limits 
      // (as a percentage of the input buffer size) that trig their transmission, 
      // and the size of the receive buffer in bytes.
   	local nArgCount
      nArgCount =  ArgCount( )
		if not this.IsPortOpen( )
			return false
		endif		
		if nArgCount >= 5
			if not this.chk( SetupComm( this.hSerial, nInBuffSize, this.CommProp.getMember( "dwCurrentTxQueue" )))
				return false
			endif
			if not this.chk( getCommProperties( this.hSerial, this.CommProp.value ))
				return false
			endif
			this.XonXoffSet.InBufSize = nInBuffSize
		endif
		with ( this.dcb )
			setMember( "XonChar", nXonChar)
			setMember( "XoffChar", nXoffChar )
			setMember( "fTXContinueOnXoff", 0 ) //iif( fTxAfterXoffSent, 1, 0 ))
			if nArgCount >= 3
				setMember( "XonLim", nXonPct * this.XonXoffSet.InBufSize / 100 )
			endif
			if nArgCount >= 4
				setMember( "XoffLim", ( 1 - nXoffPct / 100 ) * this.XonXoffSet.InBufSize )
			endif
		endwith
		if not this.chk( SetCommState( this.hSerial, this.dcb.value ))  	// Sets the new DCB values
			return false
		endif
		with ( this.XonXoffSet )
			XonChar = nXonChar
			XoffChar = nXoffChar
			// TxAfterXoffSent = fTxAfterXoffSent
			if nArgCount >= 3
				XonPercent = nXonPct
			endif
         if nArgCount >= 4
      	   XoffPercent = nXoffPct 
         endif
		endwith
		return true

	Function checkCommErrors
   	// Checks the comm errors that occured since last call and updates the CommErrors
	   // object if new ones have appeared.
		local nErrorByte
		if not this.IsPortOpen( )
			return false
		endif
		nErrorByte = 0
		if not this.chk( ClearCommError( this.hSerial, nErrorByte, this.ComStat.value ))
			return false
		endif
		if nErrorByte # 0
			this.CommErrors.type = new array( )            // Clear arrays
			this.CommErrors.text = new array( )
			if ( bitand( nErrorByte, CE_BREAK ) # 0 )
				this.CommErrors.type.add( 1 )
				this.CommErrors.text.add( "Break" )
				return false                   // Break : do not check comm errors
			endif
			if ( bitand( nErrorByte, CE_FRAME ) # 0 )
				this.CommErrors.type.add( 2 )
				this.CommErrors.text.add( "Framing" )
			endif
			if ( bitand( nErrorByte, CE_OVERRUN ) # 0 )
				this.CommErrors.type.add( 3 )
				this.CommErrors.text.add( "Char overrun" )
			endif
			if ( bitand( nErrorByte, CE_RXPARITY ) # 0 )
				this.CommErrors.type.add( 4 )
				this.CommErrors.text.add( "Parity" )
			endif
			if ( bitand( nErrorByte, CE_TXFULL ) # 0 )
				this.CommErrors.type.add( 5 )
				this.CommErrors.text.add( "Transmit overflow" )
			endif
			if ( bitand( nErrorByte, CE_RXOVER ) # 0 )
				this.CommErrors.type.add( 6 )
				this.CommErrors.text.add( "Receive overflow" )
			endif
			return false  // Error
		endif
		return true      // no errors

	Function checkPortStatus
   	// This method updates the CommErrors and PortStatus properties
	   // of the serialComm component.
		if not this.IsPortOpen( )
			return false
		endif
		if this.CheckCommErrors( )                        // No errors occured
  			this.CommErrors.type = new array( )            // so clear arrays
			this.CommErrors.text = new array( )
      endif
		with ( this.PortStatus )
			CTSHold := ( this.ComStat.getMember( "fCtsHold" ) == 1 )
			DSRHold := ( this.ComStat.getMember( "fDsrHold" ) == 1 )
			XoffHold := ( this.ComStat.getMember( "fXoffHold" ) == 1 )
			XoffSent := ( this.ComStat.getMember( "fXoffSent" ) == 1 ) 
			CharsInRxQ := this.ComStat.getMember( "cbInQue" )
      endwith
		return true

	function IsPortOpen
   	// this function returns TRUE if the port is opened, false otherwise
	   // It displays a message after trying to execute a function while the 
   	// com port is not yet opened. The message is displayed if the displayMessages 
	   // property of the component is set to true 
		if ( this.hSerial == MINUS_ONE )
			if this.displayMessages
				AlertMessage( 'Cannot execute function, the Com Port is not opened', ERRORTITLE )
			endif
			return false
		endif
		return true

endclass

// class for the CommErrors property
class CommErrorClass
	this.type = new array( )            // Types of error
	this.text = new array( )            // Error text strings
endclass

// class for the PortStatus property 
class PortStatusClass
	this.CTSHold = false						// Tx waiting for CTS signal (If hardware handshake)
	this.DSRHold = false						// Tx waiting for DSR signal (If hardware handshake)
	this.XoffHold = false					// Tx waiting, XOFF char rec'd
	this.XoffSent = false					// Tx waiting, XOFF char sent 
	this.CharsInRxQ = 0	  					// # of bytes in input buffer, not yet read
	this.CharsNotSent = 0					// # of bytes not sent after a write due to an error.
endclass

// class for the XonXoffSet property
class XonXoffConfigClass
	this.XonChar = DEFAULT_XON 			// Xon char currently set
	this.XoffChar = DEFAULT_XOFF        // Xoff char currently set
	this.XonPercent = 0						// Percentage of receive buffer size that trigs Xon
	this.XoffPercent = 0						// Percentage of receive buffer size that trigs Xoff
	this.InBufSize = Null					// Actual receive buffer size
endclass


/*------------------------------------------------------------------------------
     This section contains Windows API C++ type structure declarations 
  	  that are used for some Win 32 API calls involving communication ports. 
  	  These structure classes are inherited from a base class defined in the
  	  file StructureEx.cc
--------------------------------------------------------------------------------*/

#include <StructApi.h>
#define TYPE_DCB   TYPE_STRING

class _COMMTIMEOUTS of structureEx
	super::addMember( TYPE_DWORD,   "ReadIntervalTimeout" )
	super::addMember( TYPE_DWORD,   "ReadTotalTimeoutMultiplier" )
	super::addMember( TYPE_DWORD,   "ReadTotalTimeoutConstant" )
	super::addMember( TYPE_DWORD,   "WriteTotalTimeoutMultiplier" )
	super::addMember( TYPE_DWORD,   "WriteTotalTimeoutConstant" )
endclass

class _COMSTAT of structureEx
	super::addMember( TYPE_DWORD,   "fCtsHold", 1 )          // Tx waiting for CTS signal
	super::addMember( TYPE_DWORD,   "fDsrHold", 1 )          // Tx waiting for DSR signal 
	super::addMember( TYPE_DWORD,   "fRlsdHold", 1 )         // Tx waiting for RLSD signal
	super::addMember( TYPE_DWORD,   "fXoffHold", 1 )         // Tx waiting, XOFF char rec'd
	super::addMember( TYPE_DWORD,   "fXoffSent", 1 )         // Tx waiting, XOFF char sent
	super::addMember( TYPE_DWORD,   "fEof", 1 )              // EOF character sent
	super::addMember( TYPE_DWORD,   "fTxim", 1 )             // character waiting for Tx
	super::addMember( TYPE_DWORD,   "fReserved", 25 )        // reserved
	super::addMember( TYPE_DWORD,   "cbInQue" )              // bytes in input buffer
	super::addMember( TYPE_DWORD,   "cbOutQue" )             // bytes in output buffer
endclass

class _DCB of structureEx
	super::addMember( TYPE_DWORD,   "DCBlength" )            // sizeof(DCB) 
	super::addMember( TYPE_DWORD,   "BaudRate" )             // current baud rate
	super::addMember( TYPE_DWORD,   "fBinary", 1 )           // binary mode, no EOF check
	super::addMember( TYPE_DWORD,   "fParity", 1 )           // enable parity checking
	super::addMember( TYPE_DWORD,   "fOutxCtsFlow", 1 )      // CTS output flow control
	super::addMember( TYPE_DWORD,   "fOutxDsrFlow", 1 )      // DSR output flow control
	super::addMember( TYPE_DWORD,   "fDtrControl", 2 )       // DTR flow control type
	super::addMember( TYPE_DWORD,   "fDsrSensitivity", 1 )   // DSR sensitivity 
	super::addMember( TYPE_DWORD,   "fTXContinueOnXoff", 1 ) // XOFF continues Tx
	super::addMember( TYPE_DWORD,   "fOutX", 1 )             // XON/XOFF out flow control
	super::addMember( TYPE_DWORD,   "fInX", 1 )              // XON/XOFF in flow control
	super::addMember( TYPE_DWORD,   "fErrorChar", 1 )        // enable error replacement
	super::addMember( TYPE_DWORD,   "fNull", 1 )             // enable null stripping
	super::addMember( TYPE_DWORD,   "fRtsControl", 2 )       // RTS flow control
	super::addMember( TYPE_DWORD,   "fAbortOnError", 1 )     // abort reads/writes on error 
	super::addMember( TYPE_DWORD,   "fDummy2", 17 )          // reserved 
	super::addMember( TYPE_WORD,    "wReserved" )            // not currently used
	super::addMember( TYPE_WORD,    "XonLim" )               // transmit XON threshold
	super::addMember( TYPE_WORD,    "XoffLim" )              // transmit XOFF threshold
	super::addMember( TYPE_BYTE,    "ByteSize" )             // number of bits/byte, 4-8 
	super::addMember( TYPE_BYTE,    "Parity" )               // 0-4=no,odd,even,mark,space
	super::addMember( TYPE_BYTE,    "StopBits" )             // 0,1,2 = 1, 1.5, 2 
	super::addMember( TYPE_CHAR,    "XonChar" )              // Tx and Rx XON character
	super::addMember( TYPE_CHAR,    "XoffChar" )             // Tx and Rx XOFF character
	super::addMember( TYPE_CHAR,    "ErrorChar" )            // error replacement character
	super::addMember( TYPE_CHAR,    "EofChar" )              // end of input character 
	super::addMember( TYPE_CHAR,    "EvtChar" )              // received event character 
	super::addMember( TYPE_WORD,    "wReserved1" )           // reserved; do not use
endclass

class _COMMCONFIG of structureEx
	super::addMember( TYPE_DWORD,   "dwSize" )
	super::addMember( TYPE_WORD,    "wVersion" )
	super::addMember( TYPE_WORD,    "wReserved" )
	super::addMember( TYPE_DCB,     "dcb", new _DCB().length() )      // sizeOf( DCB )
	super::addMember( TYPE_DWORD,   "dwProviderSubType" )
	super::addMember( TYPE_DWORD,   "dwProviderOffset" )
	super::addMember( TYPE_DWORD,   "dwProviderSize" )
	super::addMember( TYPE_STRING,    "wcProviderData[1]", 128 )
endclass

class _COMMPROP of structureEx
	super::addMember( TYPE_WORD,    "wPacketLength" )       	// packet size, in bytes 
	super::addMember( TYPE_WORD,    "wPacketVersion" )      	// packet version 
	super::addMember( TYPE_DWORD,   "dwServiceMask" )      	// services implemented
	super::addMember( TYPE_DWORD,   "dwReserved1" )        	// reserved 
	super::addMember( TYPE_DWORD,   "dwMaxTxQueue" )       	// max Tx bufsize, in bytes
	super::addMember( TYPE_DWORD,   "dwMaxRxQueue" )       	// max Rx bufsize, in bytes 
	super::addMember( TYPE_DWORD,   "dwMaxBaud" )          	// max baud rate, in bps 
	super::addMember( TYPE_DWORD,   "dwProvSubType" )      	// specific provider type
	super::addMember( TYPE_DWORD,   "dwProvCapabilities" ) 	// capabilities supported 
	super::addMember( TYPE_DWORD,   "dwSettableParams" )   	// changable parameters 
	super::addMember( TYPE_DWORD,   "dwSettableBaud" )     	// allowable baud rates
	super::addMember( TYPE_WORD,    "wSettableData" )      	// allowable byte sizes
	super::addMember( TYPE_WORD,    "wSettableStopParity" ) 	// stop bits/parity allowed
	super::addMember( TYPE_DWORD,   "dwCurrentTxQueue" )    	// Tx buffer size, in bytes
	super::addMember( TYPE_DWORD,   "dwCurrentRxQueue" )     // Rx buffer size, in bytes 
	super::addMember( TYPE_DWORD,   "dwProvSpec1" )          // provider-specific data
	super::addMember( TYPE_DWORD,   "dwProvSpec2" )          // provider-specific data
	super::addMember( TYPE_WORD,    "wcProvChar[1]" )        // provider-specific data 
endclass

class _MODEMDEVCAPS of structureEx
	super::addMember( TYPE_DWORD,		"dwActualSize" )               // size of returned data, in bytes
	super::addMember( TYPE_DWORD,		"dwRequiredSize" )             // total size of structure
	super::addMember( TYPE_DWORD,		"dwDevSpecificOffset" )        // offset of provider-defined data
	super::addMember( TYPE_DWORD,		"dwDevSpecificSize" )          // size of provider-defined data

	 // Product and version identification
	super::addMember( TYPE_DWORD,		"dwModemProviderVersion" )     // provider version number
	super::addMember( TYPE_DWORD,		"dwModemManufacturerOffset" )  // offset of manufacturer name
	super::addMember( TYPE_DWORD,		"dwModemManufacturerSize" )    // length of manufacturer name
	super::addMember( TYPE_DWORD,		"dwModemModelOffset" )         // offset of model name
	super::addMember( TYPE_DWORD,		"dwModemModelSize" )           // length of model name
	super::addMember( TYPE_DWORD,		"dwModemVersionOffset" )       // offset of version name
	super::addMember( TYPE_DWORD,		"dwModemVersionSize" )         // length of version name

	 // Local option capabilities
	super::addMember( TYPE_DWORD,		"dwDialOptions" )              // bitmap of supported values
	super::addMember( TYPE_DWORD,		"dwCallSetupFailTimer" )       // maximum in seconds
	super::addMember( TYPE_DWORD,		"dwInactivityTimeout" )        // maximum in tenths of seconds
	super::addMember( TYPE_DWORD,		"dwSpeakerVolume" )            // bitmap of supported values
	super::addMember( TYPE_DWORD,		"dwSpeakerMode" )              // bitmap of supported values
	super::addMember( TYPE_DWORD,		"dwModemOptions" )             // bitmap of supported values
	super::addMember( TYPE_DWORD,		"dwMaxDTERate" )               // maximum value in bit/s
	super::addMember( TYPE_DWORD,		"dwMaxDCERate" )               // maximum value in bit/s

	 // Variable portion for strings and provider-specific data
	super::addMember( TYPE_BYTE,		"abVariablePortion[1]" )       // variable-length data
endclass

class _MODEMSETTINGS of structureEx
	super::addMember( TYPE_DWORD,		"dwActualSize" )               // size of returned data, in bytes
	super::addMember( TYPE_DWORD,		"dwRequiredSize" )             // total size of structure
	super::addMember( TYPE_DWORD,		"dwDevSpecificOffset" )        // offset of provider-defined data
	super::addMember( TYPE_DWORD,		"dwDevSpecificSize" )          // size of provider-defined data

	 // Static local options (read/write)
	super::addMember( TYPE_DWORD,		"dwCallSetupFailTimer" )       // call setup timeout, in seconds
	super::addMember( TYPE_DWORD,		"dwInactivityTimeout" )        // inactivity timeout, in tenths of seconds
	super::addMember( TYPE_DWORD,		"dwSpeakerVolume" )            // speaker volume level
	super::addMember( TYPE_DWORD,		"dwSpeakerMode" )              // speaker mode
	super::addMember( TYPE_DWORD,		"dwPreferredModemOptions" )    // bitmap specifying preferred options

	 // negotiated options (read only) for current or last call
	super::addMember( TYPE_DWORD,		"dwNegotiatedModemOptions" )   // bitmap specifying actual options
	super::addMember( TYPE_DWORD,		"dwNegotiatedDCERate" )        // DCE rate, in bits per second

	 // Variable portion for proprietary expansion
	super::addMember( TYPE_BYTE,	   "abVariablePortion[1]" )       // variable-length data
endclass


