/*
    SetupSourceAlias.prg
    A simple program designed to make it as easy as possible
    for a new user of the dUFLP to set up the Source Alias
    needed to run much of the code in this library.

    Created: December 18, 2001
    Author.: Ken Mayer

    Updated: February 9, 2002
       Bernd Hohenester pointed out after posting release H of
       the library that there were some problems:
         1) Checking for the "dUFLP" key in line 51 was a problem
            the way it was being done (fixed),
         2) INI.CC was updated after this was written to use
            the :dUFLP: Source alias, and if it didn't exist,
            how was it to find things? Errors would occur, and
            the program would fail. Instead, this program calls
            the API call necessary to write to the dB2K.ini file
            directly (now).

    Updated: February, 2002
        Second update, after final posting for release H of the
        library, this one accepts two parameters that are 
        both optional cAliasName and cPath. If used without
        either parameter, it will assume the dUFLP alias,
        and the current folder. If used with just the first
        parameter, it will use that alias, and the current
        folder.

     Updated: August, 2012
        Difficulty with current versions of dBASE (2.7 and later)
        and Windows Vista/Windows 7 (and one presumes later).
        UAC Paths throw things off, and these versions default
        to using UAC Paths. The .INI file will be in a read-only
        location in these situations. So, we need to change the 
        place we're looking. Luckily this one isn't too difficult.

     Usage:
        do SetupSourceAlias [with <cAliasName>[,<cPath>]

     Parameters:
        cAliasName = Optional
                     The name of the Source Alias you wish
                     to designate, and place into the
                     SourceAliases array used by dB2K
                     This is optional, the default is
                     "dUFLP".
        cPath      = Optional
                     Cannot be used without the first parameter.
                     If used, this is the path that will
                     be used for the Source Alias. Note that
                     it should NOT contain a trailing backslash,
                     i.e., C:\MySourceLibrary\, but if you
                     put one in, the code will remove it ...
                     If not used, the program assumes the
                     current folder.
*/

parameters cAliasName, cPath, bCalled

// Default the alias:
if pCount() == 0 or empty( argVector( 1 ) ) or argVector(1)==null
   cAliasName = "dUFLP"
endif

// Default the path
if pCount() < 2 or empty( argVector( 2 ) ) or argVector(2)==null
   cPath = set("dire")
else
   // Path was passed, check for trailing
   // backslash. If exists, remove it ...
   if right( cPath, 1 ) == "\"
      cPath = left( cPath, len( cPath ) - 1 )
   endif
endif
if pCount() < 3
   bCalled = false
endif

// We need to see if this version of dBASE is using 
// the UAC settings:
if type( "_app.useUACPaths" ) # "L"
   _app.useUACPaths = false // create custom property
endif

// check for source alias -- note that
// since not everyone enters the alias in
// the "proper" case, this code is necessary
// to check for the alias as Associative Arrays
// are case sensitive
cKey = _app.SourceAliases.firstKey
lExists = false
for i = 1 to _app.SourceAliases.count()
    if cKey.toLowerCase() == cAliasName.toLowerCase()
       lExists := true
       exit
    endif
    cKey = _app.SourceAliases.nextKey( cKey )
next

// If it does exist, is it pointing to the current
// folder? 
lOK = false
if lExists and ;
   ( _app.sourceAliases[ cKey ].toLowerCase() ==  cPath.toLowerCase() )
   lOK = true
endif

// if we're ok, then nothing to do, let's let the
// user know, and move on ...
if lOK
   msgbox( "Source Alias already exists, and"+chr(13)+;
           "is pointing to the current folder.",;
           "SetupSourceAlias: Nothing to change", 64 )
   return
endif

// Can't use INI.cc here as Source Alias used
// in the code may not exist, and if it does,
// we may be removing it and recreating it, 
// so let's just grab the actual API call that
// is used:
if type("WritePrivateProfileString") <> "FP"
   extern CLOGICAL WritePrivateProfileString( CSTRING,;
                                              CSTRING,;
                                              CSTRING,;
                                              CSTRING ) ;
          kernel32 from "WritePrivateProfileStringA"
endif

// deal with UAC!!! Where is PLUS.INI file? 

// set values to be used:
if _app.useUACPaths
   cINIFile = _app.iniFile
else
   if __version__ < 2.0
      cINIFile = _dbwinhome+"bin\dB2K.ini"
   else
      cINIFile = _dbwinhome+"bin\PLUS.ini"
   endif 
endif // _app.useUACPaths
cSection = "SourcePath"

// if it exists, but not current folder, delete it
if lExists and not lOK
   WritePrivateProfileString(cSection, cAliasName, 0, cIniFile)
endif

// write new/current values:
WritePrivateProfileString(cSection, cAliasName,;
                          cPath, cIniFile)

// Flush (write to the ini file):
WritePrivateProfileString(0, 0, 0, cIniFile)

// That should do it
if not bCalled
   if msgbox( "The source alias "+cAliasName+" has been"+chr(13)+;
              "set to "+cPath+", and is now in"+chr(13)+;
              "the "+iif( __version__ < 2, "dB2K.ini", "Plus.ini" )+;
              " file. In order for this to "+chr(13)+;
              "take effect, however, you will need to "+chr(13)+;
              "exit and restart dBASE. Selecting 'Yes' will"+chr(13)+;
              "exit and restart dBASE, but you will need to"+chr(13)+;
              "change the path back to the dUFLP folder ...", ;
              "Exit/Restart dBASE?", 36 ) == 6 // Yes
      run(true,_app.EXEName)
      quit
   endif
endif
// End of Program: setupSourceAlias.prg


